from unittest import TestCase

import numpy as np
import unittest.mock as mo

from grortir.main.pso.swarm import Swarm


class TestSwarm(TestCase):
    def setUp(self):
        self.stages = mo.Mock()
        self.process = mo.Mock()
        self.number_of_particles = 40
        self.tested_object = Swarm(self.process, self.stages,
                                   self.number_of_particles)
        self.mock_of_swarm = mo.Mock()
        self.mock_of_particle_1 = mo.Mock()
        self.mock_of_particle_2 = mo.Mock()
        self.mock_of_swarm.particles = [self.mock_of_particle_1,
                                        self.mock_of_particle_2]
        self.mock_of_best_particle = mo.Mock()
        self.mock_of_swarm.best_particle = self.mock_of_best_particle

    def test_constructor(self):
        self.assertIsNotNone(self.tested_object)
        self.assertEqual(self.stages, self.tested_object.stages)
        self.assertEqual(self.tested_object.number_of_particles,
                         self.number_of_particles)

    def test_initialize(self):
        Swarm.initialize(self.mock_of_swarm)
        self.mock_of_particle_1.initialize.assert_any_call()
        self.mock_of_particle_2.initialize.assert_any_call()

    def test_do_single_iteration(self):
        self.mock_of_swarm.particles = [self.mock_of_particle_1,
                                        self.mock_of_particle_1]
        expected_calls_particle = [mo.call.update_values(),
                                   mo.call.update_values(),
                                   mo.call.move(), mo.call.move()]
        expected_calls_swarm = [mo.call._update_best_particle(),
                                mo.call._update_velocieties()]

        Swarm.do_single_iteration(self.mock_of_swarm)

        self.assertListEqual(expected_calls_particle,
                             self.mock_of_particle_1.method_calls)
        self.assertListEqual(expected_calls_swarm,
                             self.mock_of_swarm.method_calls)

    def test__update_best_particle_1(self):
        self.mock_of_swarm.best_particle_quality = np.inf
        self.mock_of_particle_1.best_quality = 7
        self.mock_of_particle_2.best_quality = 6

        Swarm._update_best_particle(self.mock_of_swarm)

        self.assertEqual(self.mock_of_swarm.best_particle_quality, 6)
        self.assertEqual(self.mock_of_swarm.best_particle,
                         self.mock_of_particle_2)

    def test__update_best_particle_2(self):
        self.mock_of_swarm.best_particle_quality = np.inf
        self.mock_of_particle_1.best_quality = 7
        self.mock_of_particle_2.best_quality = 8

        Swarm._update_best_particle(self.mock_of_swarm)

        self.assertEqual(self.mock_of_swarm.best_particle_quality, 7)
        self.assertEqual(self.mock_of_swarm.best_particle,
                         self.mock_of_particle_1)

    def test__update_velocieties(self):
        Swarm._update_velocieties(self.mock_of_swarm)
        self.mock_of_particle_1.update_velocities.assert_called_once_with(
            self.mock_of_best_particle)
        self.mock_of_particle_2.update_velocities.assert_called_once_with(
            self.mock_of_best_particle)



"""Represents optimization strategy for group in PSO."""
import numpy as np

from grortir.main.pso.group_optimization_strategy import \
    GroupOptimizationStrategy


class CallsGroupOptimizationStrategy(GroupOptimizationStrategy):
    """Represents optimization strategy for group in PSO."""

    def __init__(self, stages_in_group):
        self.stages_in_group = stages_in_group
        self.max_cost = 0
        self.expected_quality = np.inf

    def initialize(self):
        """
            Initialize strategy.
        """
        max_cost = 0
        for stage in self.stages_in_group:
            max_cost += stage.get_maximal_acceptable_cost()
            if self.expected_quality > stage.maximum_acceptable_quality:
                self.expected_quality = stage.maximum_acceptable_quality
        self.max_cost = max_cost

    def should_continue(self, best_particle):
        """

        Args:
            best_particle Particle: best particle in swarm.

        Returns:
            bool: true if continuation is required.

        """
        return self._is_safe_cost() and not self._is_enough_quality(
            best_particle)

    def _is_safe_cost(self):
        current_cost = 0
        for stage in self.stages_in_group:
            current_cost += stage.get_cost()
        return current_cost < self.max_cost

    def _is_enough_quality(self, best_particle):
        return best_particle.best_quality <= self.expected_quality



"""Tests for pyswarm."""
from unittest import TestCase

from grortir.externals.pyswarm.pso import pso
from grortir.main.model.core.abstract_stage import AbstractStage


class TestPso(TestCase):
    """Class for tests pyswarm."""

    def test_run_simple_pso(self):
        """Test running library."""

        stage_to_test = ExampleStage(EXAMPLE_INPUT)
        x_opt, f_opt, iterations = pso(stage=stage_to_test)
        self.assertIsNotNone(x_opt)
        self.assertIsNotNone(f_opt)
        self.assertIsNotNone(iterations)

    def test_not_able_to_optimize_stage(self):
        """Test when stage is not able to optimize."""
        stage_not_able_to_optimize = StageNotAbleToOptimize(EXAMPLE_INPUT)
        x_opt, f_opt, iterations = pso(stage=stage_not_able_to_optimize)
        self.assertIsNotNone(x_opt)
        self.assertIsNotNone(f_opt)
        self.assertIsNotNone(iterations)
        self.assertEqual(iterations, 1)


# Test data:
LOWER_BOUND = [-3, -1]
UPPER_BOUND = [2, 6]
EXAMPLE_INPUT = (-1, 3)


class ExampleStage(AbstractStage):
    """Stage for testing."""

    def __init__(self, input_vector):
        super().__init__(input_vector)
        self.lower_bounds = LOWER_BOUND
        self.upper_bounds = UPPER_BOUND
        self.counter = 0

    @staticmethod
    def get_cost():
        """No costs."""
        return 0

    def get_quality(self, vector=None):
        """Return quality."""
        self.control_params = vector
        self.counter += 1
        return myfunc(self.control_params)

    @staticmethod
    def could_be_optimized():
        """Return True."""
        return True

    def is_enough_quality(self, value):
        """Return False."""
        return False

    @staticmethod
    def get_output_of_stage():
        """Return None."""
        return None


class StageNotAbleToOptimize(ExampleStage):
    """Stage which couldn't be optimized."""

    @staticmethod
    def could_be_optimized():
        """Returns false."""
        return False

    @staticmethod
    def get_output_of_stage():
        """Return None."""
        return None


def myfunc(input_vector):
    """Simple function for tests.

    Args:
        input_vector (list): input vector

    Returns:
        float: value of function
    """
    x_1 = input_vector[0]
    x_2 = input_vector[1]
    return x_1 ** 4 - 2 * x_2 * x_1 ** 2 + x_2 ** 2 + x_1 ** 2 - 2 * x_1 + 5



"""Validate process."""


class ProcessValidator(object):
    """Validate if process is correctly defined."""
    @staticmethod
    def validate(process):
        """Main method."""
        return process is not None



"""Module represent single stage in process."""
import numpy as np

from grortir.main.model.core.optimization_status import OptimizationStatus


# pylint: disable=too-many-instance-attributes


class AbstractStage:
    """Class represent single stage in process.

    Attributes:
        input_vector (list): initial vector
        optimization_status (OptimizationStatus): status of optimization
        control_params (list): actual control params
        lower_bounds (list): lower bounds for control params
        upper_bounds (list): upper bounds for control params

    """

    def __init__(self, input_vector=()):
        """Constructor.

        Args:
            input_vector (tuple): initial vector
        """
        self.input_vector = input_vector
        self.optimization_status = OptimizationStatus.not_started
        self.control_params = np.zeros_like(input_vector)
        dimensions = len(input_vector)
        self.lower_bounds = [-1] * dimensions
        self.upper_bounds = [1] * dimensions
        self.final_output = None
        self.final_quality = None
        self.final_cost = None

    @staticmethod
    def get_cost():
        """Return cost of stage."""
        raise NotImplementedError

    @staticmethod
    def get_quality(input_vector=None, control_params=None):
        """Return quality of stage."""
        raise NotImplementedError

    @staticmethod
    def could_be_optimized():
        """Return answer if it is still possible to optimize that stage."""
        raise NotImplementedError

    @staticmethod
    def is_enough_quality(value):
        """Return True if value is enough quality."""
        raise NotImplementedError

    @staticmethod
    def get_output_of_stage():
        """Result of processing input with current control params."""
        raise NotImplementedError

    @staticmethod
    def get_maximal_acceptable_cost():
        """Return maximum acceptable cost."""
        raise NotImplementedError



from unittest import TestCase
from unittest.mock import Mock

from grortir.main.optimizers.grouping_strategy import GroupingStrategy


class TestGroupingStrategy(TestCase):
    def test_get_items_from_group(self):
        grouping_strategy = Mock()
        grouping_strategy.get_actual_numbers_of_groups.return_value = 3
        with self.assertRaises(ValueError):
            GroupingStrategy.get_items_from_group(grouping_strategy, 4)



"""Basic optimizer."""

from grortir.externals.pyswarm.pso import pso
from grortir.main.model.core.optimization_status import OptimizationStatus
from grortir.main.optimizers.base_optimizer import BaseOptimizer


class Optimizer(BaseOptimizer):
    """Optimizer is object which optimize process."""

    def optimize_process(self):
        """Optimize process.
        Raises:
            AttributeError: When project has incorrect structure.

        Returns:
            bool: True if success, False otherwise."""
        for stage in self.ordered_stages:
            if not self.process.predecessors(stage):
                self.run_pso(stage)
            elif len(self.process.predecessors(stage)) == 1:
                predecessor = self.process.predecessors(stage)[0]
                stage.input_vector = predecessor.get_output_of_stage(
                    predecessor.input_vector, predecessor.control_params)
                self.run_pso(stage)
            else:
                raise AttributeError('Incorrect process structure.')

            if stage.optimization_status != OptimizationStatus.success:
                return False
        return True

    def run_pso(self, stage):
        """Run pso with predefined parameters."""
        pso(stage, swarmsize=self.swarm_size)



"""Representation of a particle in swarm."""
# pylint: disable=too-many-instance-attributes
# pylint: disable=redefined-variable-type
from collections import OrderedDict

import numpy as np

from grortir.main.model.core.optimization_status import OptimizationStatus
from grortir.main.pso.position_updater import PositionUpdater
from grortir.main.pso.velocity_calculator import VelocityCalculator


class Particle(object):
    """Implementation of particle."""

    def __init__(self, stages, process, number):
        self.stages = stages
        self.process = process
        self.number = number
        self.velocity_calculator = VelocityCalculator()
        self.current_velocities = {}
        self.current_quality = {}
        self.best_quality = np.inf
        self.best_positions = {}
        self.input_vectors_for_best_pos = {}
        self.current_control_params = OrderedDict()
        self.position_updater = PositionUpdater(self.current_control_params)
        self.current_input = {}

    def initialize(self):
        """Initialization of single particle."""
        for stage in self.stages:
            self.current_control_params[stage] = stage.control_params
            self.current_input[stage] = stage.input_vector

            stage.optimization_status = OptimizationStatus.in_progress
            self.current_quality[stage] = np.inf
        self._set_initial_positions()
        self._set_initial_velocities()

    def _set_initial_positions(self):
        self.current_control_params = self.position_updater. \
            set_initial_control_params(self.current_control_params)

    def _set_initial_velocities(self):
        self.current_velocities = self.velocity_calculator. \
            calculate_initial_velocity(self.current_control_params)

    def update_values(self):
        """Update values in swarm."""
        self.update_input_vectors()
        self.calculate_current_quality()
        self._update_stages_status()
        self._update_best_position()

    def _update_best_position(self):
        current_quality = self.get_the_overall_quality()
        if current_quality < self.best_quality:
            self.best_quality = current_quality
            for stage in self.stages:
                self.best_positions[stage] = self.current_control_params[stage]
                self.input_vectors_for_best_pos[stage] = self.current_input[
                    stage]

    def update_input_vectors(self):
        """Update input vectors in all stages."""
        for stage in self.stages:
            current_output = stage.get_output_of_stage(
                self.current_input[stage], self.current_control_params[stage])
            successors = self.process.successors(stage)
            for successor in successors:
                successor.input_vector = current_output
                self.current_input[successor] = current_output

    def calculate_current_quality(self):
        """Calculate current quality."""
        for stage in self.stages:
            self.current_quality[stage] = stage.get_quality(
                self.current_input[stage],
                self.current_control_params[stage])

    def get_the_overall_quality(self):
        """Return overall quality of stages."""
        stage = max(self.current_quality, key=self.current_quality.get)
        return self.current_quality[stage]

    def move(self):
        """Move particle."""
        self.current_control_params = self.position_updater.update_position(
            self.current_velocities, self.current_control_params)

    def update_velocities(self, best_particle):
        """Update velocities in swarm."""
        velocity = self.velocity_calculator.calculate(
            self.current_velocities,
            self.best_positions,
            best_particle.best_positions,
            self.current_control_params)
        self.current_velocities = velocity

    def _update_stages_status(self):
        for stage in self.stages:
            if stage.is_enough_quality(self.current_quality[stage]):
                stage.optimization_status = OptimizationStatus.success
            if not stage.could_be_optimized():
                stage.optimization_status = OptimizationStatus.failed



"""Test class for module calls_stage."""

from unittest import TestCase
from unittest.mock import Mock, sentinel

from grortir.main.model.stages.calls_stage import CallsStage

MAX_CALLS = 100
CONTRL_PARAMS = [1, 1, 1, 1, 1, 1.5]


class TestCallsStage(TestCase):
    """Test class for CallsStage."""

    def test_get_quality(self):
        """Test for get_quality method."""
        tested_object = Mock()
        tested_object.cost = 7
        tested_object.calculate_quality.return_value = sentinel.quality
        tested_object.control_params = CONTRL_PARAMS
        result = CallsStage.get_quality(tested_object, CONTRL_PARAMS)
        self.assertEqual(tested_object.cost, 8)
        tested_object.calculate_quality.assert_called_with(CONTRL_PARAMS,
                                                           CONTRL_PARAMS)
        self.assertEqual(result, sentinel.quality)

    def test_get_quality_without_vector(self):
        """Test for get_quality method when no arguments passed."""
        tested_object = Mock()
        tested_object.cost = 9
        tested_object.calculate_quality.return_value = sentinel.quality
        tested_object.control_params = CONTRL_PARAMS
        CallsStage.get_quality(tested_object)
        self.assertEqual(tested_object.cost, 10)
        tested_object.calculate_quality.assert_called_with(None, CONTRL_PARAMS)

    def test_get_cost(self):
        """Test for get_cost method."""
        tested_object = Mock()
        tested_object.cost = sentinel.cost
        result = CallsStage.get_cost(tested_object)
        self.assertEqual(result, sentinel.cost)

    def test_calculate_quality_ex(self):
        """Test case when control are wrong."""
        input_vector = (2, 3, 4, 5, 6)
        tested_object = CallsStage('name', MAX_CALLS, input_vector)
        tested_object.control_params = [2, 2]
        with self.assertRaises(AssertionError):
            tested_object.calculate_quality(input_vector,
                                            tested_object.control_params)

    def test_calculate_quality_ok(self):
        """Test case when control params and input are okay."""
        input_vector = (2, 3, 4, 5, 6, 1)
        tested_object = CallsStage('name', MAX_CALLS, input_vector)
        tested_object.control_params = CONTRL_PARAMS
        result = tested_object.calculate_quality(input_vector,
                                                 tested_object.control_params)
        self.assertEqual(result, 55.25)

    def test_could_be_optimized_pos(self):
        """Positive case for test could_be_optimized method."""
        tested_object = Mock()
        tested_object.get_cost.return_value = MAX_CALLS - 1
        tested_object.get_maximal_acceptable_cost.return_value = MAX_CALLS
        result = CallsStage.could_be_optimized(tested_object)
        self.assertTrue(result)

    def test_could_be_optimized_neg(self):
        """Negative case for test could_be_optimized method."""
        tested_object = Mock()
        tested_object.get_cost.return_value = MAX_CALLS + 1
        tested_object.get_maximal_acceptable_cost.return_value = MAX_CALLS
        result = CallsStage.could_be_optimized(tested_object)
        self.assertFalse(result)

    def test_is_enough_quality(self):
        """Test is_enough_quality method."""
        tested_object = Mock()
        tested_object.maximum_acceptable_quality = 7
        result = CallsStage.is_enough_quality(tested_object, 8)
        self.assertFalse(result)

    def test_get_output_of_stage(self):
        """Test returning output."""
        tested_object = CallsStage('name', MAX_CALLS)
        result = tested_object.get_output_of_stage([], [])
        self.assertEqual([], result)

    def test_get_maximal_acceptable_cost(self):
        tested_object = CallsStage('name', MAX_CALLS)
        result = tested_object.get_maximal_acceptable_cost()
        self.assertEqual(result, MAX_CALLS)



from tests.framework.calls.calls_record import CallRecord


# pylint: disable=too-many-arguments


class CallsRegistry:
    def __init__(self):
        self.calls_storage = {}
        self.calls_counter = 0

    def register_call(self, call_record):
        """
        Register call.
        Args:
            call_record (CallRecord): represent call
        """
        self.calls_counter += 1
        if self.calls_storage.get(call_record.entity_name) is None:
            self.calls_storage[call_record.entity_name] = {}
        if self.calls_storage[call_record.entity_name].get(
                call_record.method_name) is None:
            self.calls_storage[call_record.entity_name][
                call_record.method_name] = []
        self.calls_storage[call_record.entity_name][
            call_record.method_name].append({"call": call_record,
                                             "calls_count": self.calls_counter})

    def get_method_calls(self, entity_name, method_name):
        """
        Return calls
        Args:
            entity_name (str): name of entity
            method_name (str): name of method called on entity

        Returns:
            list: list of call objects  (CallRecord)

        """
        return self.calls_storage[entity_name][method_name]

    def add_call(self, entity_name, method_name, input_parms_list, output,
                 count):
        """Add and register call of function.

        Args:
            entity_name (str): name of entity
            method_name (str): name of method
            input_parms_list (list): list of arguments put to function
            output (object):  returned value by method_name
            count (int): number of count
        """
        call_record = CallRecord(entity_name, method_name, input_parms_list,
                                 output, count)
        self.register_call(call_record)



"""Contain mechanism for changing position."""

import numpy as np


class PositionUpdater:
    """Class responsible for moving objects.

        Attributes:
            stage (AbstractStage): Stage in which we are going to move.
            control_params (dict): Control params for stages.
    """

    def __init__(self, control_params):
        self.control_params = control_params

    @staticmethod
    def set_initial_control_params(control_params):
        """Set initial positions."""
        for stage in control_params:
            random = np.random.rand(len(control_params[stage]))
            delta = np.asarray(stage.upper_bounds) - np.asarray(
                stage.lower_bounds)
            new_control_params = np.asarray(stage.lower_bounds) + random * delta
            control_params[stage] = new_control_params.tolist()
        return control_params

    def update_position(self, velocities, control_params):
        """Update positions."""
        for stage in velocities:
            new_control_params = control_params[stage] + velocities[stage]
            control_params[stage] = self._fix_coordinates(stage,
                                                          new_control_params)
        return control_params

    @staticmethod
    def _fix_coordinates(stage, control_params):
        for index, single_param in enumerate(control_params):
            if single_param > stage.upper_bounds[index]:
                control_params[index] = stage.upper_bounds[index]
            elif single_param < stage.lower_bounds[index]:
                control_params[index] = stage.lower_bounds[index]
        return control_params



"""Contains factory class of process with calls stages."""

from grortir.main.model.processes.calls_process import CallsProcess
from grortir.main.model.stages.calls_stage import CallsStage


class CallsProcessFactory:
    """Class which produce process with the same type of stages.

    Attributes:
        structure_type (str): represent type of structure
        how_many_nodes (int): how many nodes should be in generated process
        max_calls (int): max_calls for stages
        process (CallsProcess): process which will be incrementally build
        initial_vec (tuple): initial vector
    """

    def __init__(self, structure_type, how_many_nodes, max_calls,
                 initial_vec=()):
        """Constructor."""
        self.max_calls = max_calls
        self._stages_to_add = []
        self.structure_type = structure_type
        self.how_many_nodes = how_many_nodes
        self.process = CallsProcess()
        self.initial_vec = initial_vec

    def _update_structure(self):
        """Update structure of process."""
        if self.structure_type == "linear":
            if len(self._stages_to_add) > 1:
                self.process.add_path(self._stages_to_add)
            else:
                self.process.add_node(self._stages_to_add[0])
        else:
            raise NotImplementedError()

    def _create_stages(self):
        """Create stages which will be injected to process."""
        for i in range(self.how_many_nodes):
            self._stages_to_add += [
                CallsStage(str(i), self.max_calls, self.initial_vec)]

    def construct_process(self):
        """Construct process."""
        self._create_stages()
        self._update_structure()
        return self.process



from unittest import TestCase
from unittest.mock import Mock

import numpy as np

from grortir.main.pso.particle import Particle
from grortir.main.pso.position_updater import PositionUpdater


class TestParticle(TestCase):
    def setUp(self):
        self.particle_mock = Mock()
        self.stages = [Mock(), Mock()]
        self.position_updater = Mock()
        self.velocity_calculator = Mock()
        for i in range(len(self.stages)):
            self.velocity_calculator. \
                calculate_initial_velocity.return_value = 0.01 * (i + 1)
        self.particle_mock.stages = self.stages
        self.particle_mock.position_updater = self.position_updater
        self.particle_mock.velocity_calculator = self.velocity_calculator
        self.particle_mock.current_velocities = {}
        self.particle_mock.current_quality = {stage: 100 for stage in
                                              self.stages}
        self.particle_mock.current_control_params = {}
        self.particle_mock.current_input = {}
        self.process = Mock()
        self.particle_mock.process = self.process

    def test_get_the_overall_quality_po(self):
        particle_mock = Mock()
        particle_mock.current_quality = {'a': 2, 'b': 3, 'c': -3.1234}
        result = Particle.get_the_overall_quality(particle_mock)
        self.assertEqual(result, 3)

    def test___init__(self):
        particle = Particle(self.stages, self.process, 7)
        self.assertIsInstance(particle.position_updater,
                              PositionUpdater)
        self.assertIsNotNone(particle)
        self.assertEqual(particle.best_quality, np.inf)

    def test_initialize(self):
        Particle.initialize(self.particle_mock)
        self.particle_mock._set_initial_positions.assart_any_call()
        self.particle_mock._set_initial_velocities.assart_any_call()

    def test_update_values(self):
        Particle.update_values(self.particle_mock)
        self.particle_mock.update_input_vectors.assert_any_call()
        self.particle_mock.calculate_current_quality.assert_any_call()
        self.particle_mock._update_best_position.assert_any_call()



"""Module which represent process which can be optimized."""

import networkx as nx


class AbstractProcess(nx.DiGraph):
    """Class which represent process which can be optimized."""

    def __init__(self):
        """Constructor."""
        nx.DiGraph.__init__(self)



"""Sample integration test module."""
# pylint: disable=no-self-use,missing-docstring

import unittest

from grortir import sample


class TestGrortir(unittest.TestCase):

    """Sample integration test class."""

    @staticmethod
    def test_network_stuff():
        """Example Test method."""
        assert sample.function_with_network_stuff() is True

    @staticmethod
    def test_disk_stuff():
        """Another example Test method."""
        assert sample.function_with_disk_stuff() is False



from unittest import TestCase
from unittest.mock import Mock

from grortir.main.model.core.optimization_status import OptimizationStatus
from grortir.main.optimizers.result import Result


class TestResult(TestCase):
    def setUp(self):
        self.success_stage = Mock()
        self.success_stage.optimization_status = OptimizationStatus.success
        self.failed_stage = Mock()
        self.failed_stage.optimization_status = OptimizationStatus.failed
        self.process = Mock()

    def test_generate_failed(self):
        tested_object = Result(self.process)
        self.process.nodes.return_value = [self.success_stage,
                                           self.failed_stage]
        tested_object.generate()
        self.assertEqual(tested_object.summary_result,
                         OptimizationStatus.failed)

    def test_generate_success(self):
        tested_object = Result(self.process)
        self.process.nodes.return_value = [self.success_stage,
                                           self.success_stage]
        tested_object.generate()
        self.assertEqual(tested_object.summary_result,
                         OptimizationStatus.success)



"""Basic grouped optimizer."""

from grortir.main.optimizers.result import Result


class GroupedOptimizer:
    """Optimizer is object which optimize process."""

    def __init__(self, process, grouping_strategy, pso):
        self.process = process
        self.grouping_strategy = grouping_strategy
        self.pso = pso
        self.result = Result(self.process)

    def optimize_process(self):
        """Optimize process.
        Returns:
            True if success, False otherwise."""
        self.pso.run()
        self.result.generate()



"""Represents result."""
# pylint: disable=redefined-variable-type
from grortir.main.model.core.optimization_status import OptimizationStatus


class Result(object):
    """Represents result.

        Attributes:
        process (AbstarctProcess): optimized process
    """

    def __init__(self, process):
        self.process = process
        self.summary_result = OptimizationStatus.not_started
        self.detailed_result = {}

    def generate(self):
        """Generate process optimization result."""
        for stage in self.process.nodes():
            self.detailed_result[str(stage)] = stage.optimization_status
        for stage in self.process.nodes():
            if stage.optimization_status != OptimizationStatus.success:
                self.summary_result = OptimizationStatus.failed
                return
        self.summary_result = OptimizationStatus.success



"""Module for testing grouped optimizer."""

import unittest.mock as mock
from unittest import TestCase

from grortir.main.model.core.abstract_process import AbstractProcess
from grortir.main.model.core.abstract_stage import AbstractStage
from grortir.main.optimizers.grouped_optimizer import GroupedOptimizer


class TestGroupedOptimizer(TestCase):
    """Class for testing Optimizer."""

    def setUp(self):
        """Set up environment."""
        self.some_process = AbstractProcess()
        self.first_stage = AbstractStage()
        self.second_stage = AbstractStage()
        self.third_stage_a = mock.Mock()
        self.third_stage_b = mock.Mock()
        self.some_process.add_path([self.first_stage, self.second_stage,
                                    self.third_stage_a])
        self.some_process.add_edge(self.second_stage, self.third_stage_b)

    def test___init__(self):
        """Testing creating object."""
        optimizer = GroupedOptimizer(self.some_process, mock.Mock(),
                                     mock.Mock())
        self.assertIsNotNone(optimizer)
        self.assertEqual(optimizer.process, self.some_process)



from unittest import TestCase
from unittest.mock import Mock

from grortir.main.model.core.optimization_status import OptimizationStatus
from grortir.main.pso.optimization_controller import OptimizationController


class TestOptimizationController(TestCase):
    def setUp(self):
        self.stage_success = Mock()
        self.stage_success.optimization_status = OptimizationStatus.success
        self.stage_failed = Mock()
        self.stage_failed.optimization_status = OptimizationStatus.failed
        self.stage_in_progress = Mock()
        self.stage_in_progress.optimization_status = OptimizationStatus.in_progress

    def test_should_stop_when_one_failed(self):
        self.assertFalse(OptimizationController.should_continue(
            [self.stage_success, self.stage_failed]))

    def test_should_continue_when_no_failed(self):
        self.assertTrue(OptimizationController.should_continue(
            [self.stage_success, self.stage_in_progress]))

    def test_should_stop_when_all_success(self):
        self.assertFalse((OptimizationController.should_continue(
            [self.stage_success, self.stage_success])))



"""Module for testing."""

from unittest import TestCase
from unittest.mock import Mock

from grortir.main.model.core.abstract_stage import AbstractStage


class TestAbstractStage(TestCase):
    """Class for testing AbstractStage."""

    def test___init__(self):
        """Constructor test."""
        tested_object = AbstractStage()
        self.assertIsNotNone(tested_object)

    def test___init__arguments_passed(self):
        """Check arguments in constructor."""
        abstract_stage_mock = Mock()
        input_vector = (1, 2, 3)
        AbstractStage.__init__(abstract_stage_mock, input_vector)
        self.assertEqual(abstract_stage_mock.input_vector, input_vector)

    def test___init__output_mutability(self):
        """Check fields in object."""
        input_vector = [1, 2, 3]
        tested_object = AbstractStage(input_vector)
        tested_object.input_vector[2] += 1
        self.assertEqual(tested_object.input_vector[2], 4)

    def test_get_cost(self):
        """Check get_cost method."""
        tested_object = AbstractStage()
        with self.assertRaises(NotImplementedError):
            tested_object.get_cost()

    def test_get_quality(self):
        """Check get_quality method."""
        tested_object = AbstractStage()
        with self.assertRaises(NotImplementedError):
            tested_object.get_quality()

    def test_could_be_optimized(self):
        """Check could_be_optimized method."""
        tested_object = AbstractStage()
        with self.assertRaises(NotImplementedError):
            tested_object.could_be_optimized()

    def test_is_enough_quality(self):
        """Check is_enough_quality method."""
        tested_object = AbstractStage()
        with self.assertRaises(NotImplementedError):
            tested_object.is_enough_quality(7)

    def test_get_output_of_stage(self):
        """Check is_enough_quality method."""
        tested_object = AbstractStage()
        with self.assertRaises(NotImplementedError):
            tested_object.get_output_of_stage()

    def test_get_maximal_acceptable_cost(self):
        tested_object = AbstractStage()
        with self.assertRaises(NotImplementedError):
            tested_object.get_maximal_acceptable_cost()



"""Represents swarm."""
import numpy as np

from grortir.main.pso.particle import Particle


class Swarm(object):
    """Class which represent swarm."""

    def __init__(self, process, stages, number_of_particles):
        self.stages = stages
        self.process = process
        self.number_of_particles = number_of_particles
        self.particles = [Particle(stages, self.process, i) for i in
                          range(number_of_particles)]
        self.best_particle_quality = np.inf
        self.best_particle = self.particles[0]

    def initialize(self):
        """Initialize all particles in swarm."""
        for particle in self.particles:
            particle.initialize()

    def do_single_iteration(self):
        """Iterate one time."""
        for particle in self.particles:
            particle.update_values()
        self._update_best_particle()
        self._update_velocieties()
        for particle in self.particles:
            particle.move()

    def _update_best_particle(self):
        for particle in self.particles:
            if particle.best_quality < self.best_particle_quality:
                self.best_particle = particle
                self.best_particle_quality = particle.best_quality

    def _update_velocieties(self):
        for particle in self.particles:
            particle.update_velocities(self.best_particle)

    def post_processing(self):
        """Method which should be done after all iterations."""
        best_control_params = self.best_particle.best_positions
        # calculate output:
        for stage in self.stages:
            current_output = stage.get_output_of_stage(
                stage.input_vector, best_control_params[stage])
            successors = self.process.successors(stage)
            for successor in successors:
                successor.input_vector = current_output
            stage.final_output = current_output
            stage.final_cost = stage.get_cost()
            stage.final_quality = stage.get_quality(
                stage.input_vector, best_control_params[stage])



"""Contains WholeGroupPso class."""

from grortir.main.pso.swarm import Swarm


class WholeGroupPso(object):
    """Optimize whole group."""

    def __init__(self, process, number_of_particles):
        self.process = process
        self.number_of_particles = number_of_particles

    def optimize(self, ordered_stages_to_optimize, group_optimization_strategy):
        """Optimize whole group of stages.

         Parameters:
            group_optimization_strategy (GroupOptimizationStrategy):
                strategy for optimizing groups
            ordered_stages_to_optimize (list): List of stages which need
                 to be optimized
        """
        swarm = Swarm(self.process, ordered_stages_to_optimize,
                      self.number_of_particles)
        swarm.initialize()
        group_optimization_strategy.initialize()
        while group_optimization_strategy.should_continue(swarm.best_particle):
            swarm.do_single_iteration()
        swarm.post_processing()
        print("Optimization for group finished.")



"""Package to test calls_process_factory module."""

from unittest import TestCase

from grortir.main.model.processes.calls_process import CallsProcess
from grortir.main.model.processes.factories.calls_process_factory import \
    CallsProcessFactory

MAX_CALLS = 1000


class TestCallsProcessFactory(TestCase):
    """Class to test CallsProcessFactory."""

    def test_construct_process_linear(self):
        """Test linear process construction."""
        tested_object = CallsProcessFactory("linear", 7, MAX_CALLS)
        result = tested_object.construct_process()
        self.assertIsInstance(result, CallsProcess)

    def test_construct_process_not_ex(self):
        """Test case when structure not implemented."""
        tested_object = CallsProcessFactory("not_existed", 7, MAX_CALLS)
        with self.assertRaises(NotImplementedError):
            tested_object.construct_process()



"""Testing grouping strategy."""
from unittest import TestCase
from unittest.mock import Mock

from grortir.main.optimizers.grouping_strategy import GroupingStrategy


class TestGroupingStrategy(TestCase):
    def setUp(self):
        self.stage_1 = Mock()
        self.stage_2 = Mock()
        self.stage_3 = Mock()
        self.ordered_stages = [self.stage_1, self.stage_2, self.stage_3]
        self.tested_object = GroupingStrategy(self.ordered_stages)

    def test___init__(self):
        grouping_strategy = GroupingStrategy([])
        self.assertIsNotNone(grouping_strategy)

    def test_define_group_success(self):
        self.perform_simple_grouping()
        self.assertEqual(self.tested_object.groups[self.stage_1], 0)
        self.assertEqual(self.tested_object.groups[self.stage_2], 1)
        self.assertEqual(self.tested_object.groups[self.stage_3], 1)

    def perform_simple_grouping(self):
        self.tested_object.define_group((self.stage_1,))
        self.tested_object.define_group((self.stage_2, self.stage_3))

    def test_define_group_failed(self):
        self.tested_object.define_group((self.stage_1,))
        with self.assertRaisesRegex(ValueError, "Stage already added."):
            self.tested_object.define_group((self.stage_1,))

    def test_define_group_failed_not_cons(self):
        with self.assertRaisesRegex(ValueError,
                                    "Stage not considered in strategy."):
            self.tested_object.define_group((Mock(),))

    def test_get_actual_number_of_groups_0(self):
        tested_strategy = GroupingStrategy([self.stage_1, self.stage_2])
        result = tested_strategy.get_actual_numbers_of_groups()
        self.assertEqual(result, 0)

    def test_get_actual_number_of_groups_2(self):
        self.perform_simple_grouping()
        result = self.tested_object.get_actual_numbers_of_groups()
        self.assertEqual(result, 2)

    def test_get_items_from_the_same_group_1(self):
        self.perform_simple_grouping()
        result = self.tested_object.get_items_from_the_same_group(self.stage_1)
        self.assertListEqual(result, [self.stage_1])

    def test_get_items_from_the_same_group_2(self):
        self.perform_simple_grouping()
        result = self.tested_object.get_items_from_the_same_group(self.stage_2)
        self.assertEqual(set(result), set([self.stage_2, self.stage_3]))

    def test_get_items_from_the_same_group_0(self):
        with self.assertRaises(ValueError):
            self.tested_object.get_items_from_the_same_group(Mock())



"""Class represents grouping strategy."""
STAGE_NOT_IN_STRATEGY_ = "Stage not considered in strategy."


class GroupingStrategy(object):
    """Class represents grouping strategy.
    Attributes:
        ordred_stages (list): List of stages which represent order in strategy.
        groups (dict): Contains information about groups.
                        Key is stage and value is number of group.
    """

    def __init__(self, ordered_stages):
        """Constructor."""
        self.ordered_stages = ordered_stages
        self.groups = {}

    def define_group(self, group_of_stages):
        """Add group of stages. Those stages will be in the same group.

        Raises:
            ValueError: If stage couldn't be added to new group.

        Args:
            group_of_stages (tuple): group of stages"""
        group_number = self.get_actual_numbers_of_groups()
        for stage in group_of_stages:
            if stage in self.groups.keys():
                raise ValueError("Stage already added.")
            elif stage not in self.ordered_stages:
                raise ValueError("Stage not considered in strategy.")
            else:
                self.groups[stage] = group_number

    def get_actual_numbers_of_groups(self):
        """Returns how many groups already.

        Returns:
            int: amount of groups
            """
        if len(self.groups) == 0:
            return 0
        else:
            return sorted(self.groups.values())[-1] + 1

    def get_items_from_the_same_group(self, stage):
        """Returns list of stages from group in which is already stage.
        Returns:
            list: list of stages from group ordered as in self.ordered_stages

        Raises:
            ValueError: If stage is not in strategy.

        Args:
            stage (AbstractStage): stage"""
        if stage not in self.groups:
            raise ValueError(STAGE_NOT_IN_STRATEGY_)
        group_number = self.groups[stage]
        stages_from_group = []
        for cur_stage in self.ordered_stages:
            if self.groups[cur_stage] == group_number:
                stages_from_group.append(cur_stage)
        return stages_from_group

    def get_items_from_group(self, group_number):
        """Returns list of stages from group with 'group_number'.
            Returns:
                list: list of stages from group ordered as in
                 self.ordered_stages

            Raises:
                ValueError: If wrong number of groups.

            Args:
                group_number (int): number of group"""
        if group_number not in range(0, self.get_actual_numbers_of_groups()):
            raise ValueError
        stages_to_return = []
        for current_stage in self.ordered_stages:
            if self.groups[current_stage] == group_number:
                stages_to_return.append(current_stage)
        return stages_to_return



"""Validator for input."""
from grortir.main.model.core.abstract_stage import AbstractStage
from tests.framework.calls.calls_registry import CallsRegistry


class InputValidator:
    """
        Attributes:
            calls_registry (CallsRegistry): calls registry
    """

    def __init__(self, calls_registry: CallsRegistry):
        """

        Args:
            calls_registry (CallsRegistry): registry of calls
        """
        self.calls_registry = calls_registry

    def validate_input(self, grouping_strategy, process):
        """
        Validate if input in stage successors are the same as output in stage.

        Args:
            grouping_strategy (GroupingStrategy): grouping strategy
            process (AbstractProcess): validated process
        Returns:
            bool: True if inputs are valid
        """
        all_stages = process.nodes()
        for stage in all_stages:
            groupies = grouping_strategy.get_items_from_the_same_group(stage)
            for successor in process.successors(stage):
                if successor in groupies:
                    result = self._validate_for_groupies(stage, successor)
                else:
                    result = self._validate_for_non_groupies(stage, successor)
                if result is False:
                    return False
        return True

    def _validate_for_groupies(self, stage, successor):
        stage_outputs = self.calls_registry. \
            get_method_calls(stage.name, "get_output_of_stage")
        successor_inputs = self.calls_registry. \
            get_method_calls(successor.name, "get_quality")
        output_values = []
        inputs = []
        for stage_output in stage_outputs:
            output_values.append(stage_output['call'].output)

        for successor_input in successor_inputs:
            inputs.append(successor_input['call'].input_parms_list[0])
        print("aa")
        return output_values == inputs

    def _validate_for_non_groupies(self, stage: AbstractStage, successor):
        stage_outputs = self.calls_registry. \
            get_method_calls(stage.name, "get_output_of_stage")
        successor_inputs = self.calls_registry. \
            get_method_calls(successor.name, "get_quality")
        print("bb")
        if len(stage_outputs) == 0:
            return len(successor_inputs) == 0
        else:
            last_output = stage_outputs[-1]['call'].output
            first_input = successor_inputs[0]['call'].input_parms_list[0]
            return last_output == first_input



"""Contains things related to calculating velocity."""

import numpy as np


class VelocityCalculator:
    """Calculate Velocity for a single stage.

    Attributes:
        current_velocity (list) : list of coordinates of speed for current stage
        """

    def __init__(self):
        self.c_1 = 1
        self.c_2 = 1
        self.w_factor = 0.8
        self.current_velocity = None

    @staticmethod
    def calculate_initial_velocity(control_params):
        """Calculate initial velocity."""
        velocities = {}
        for stage in control_params:
            random_factor = np.random.rand(len(control_params[stage]))
            velocity = 0.02 * random_factor - 0.01
            velocities[stage] = velocity
        return velocities

    def calculate(self, current_velocities, particle_best_positions,
                  leader_best_positions,
                  control_params):
        """Calculate velocity."""
        rand_1 = np.random.rand()
        rand_2 = np.random.rand()
        velocities = {}
        for stage in control_params:
            s_0 = self._s0(current_velocities[stage])
            s_1 = self.c_1 * self._s1(rand_1, leader_best_positions[stage],
                                      control_params[stage])
            s_2 = self.c_2 * self._s2(rand_2, particle_best_positions[stage],
                                      control_params[stage])
            velocity = s_0 + s_1 + s_2
            velocities[stage] = velocity
        return velocities

    @staticmethod
    def _s2(random_factor, particle_best_position, control_params):
        return random_factor * (
            np.asarray(particle_best_position) - np.asarray(
                control_params))

    @staticmethod
    def _s1(random_factor, leader_best_position, control_params):
        return random_factor * (
            np.asarray(leader_best_position) - np.asarray(
                control_params))

    def _s0(self, current_velocity):
        return self.w_factor * current_velocity



"""Represents optimization strategy for PSO."""
from grortir.main.pso.calls_group_optimization_strategy import \
    CallsGroupOptimizationStrategy
from grortir.main.pso.optimization_strategy import OptimizationStrategy


class CallsOptimizationStrategy(OptimizationStrategy):
    """Represents optimization strategy Calls stages for PSO."""

    def get_group_optimization_strategy(self, stages_in_group):
        """
            Return group optimization strategy.

        Args:
            stages_in_group (list): list of stages

        Returns:
            CallsGroupOptimizationStrategy: strategy for group optimization.
        """
        return CallsGroupOptimizationStrategy(stages_in_group)



"""Contains class of Calls Process."""
from grortir.main.model.core.abstract_process import AbstractProcess


class CallsProcess(AbstractProcess):
    """Implementation of process with the same type of stages - calls."""



# pylint: skip-file
from functools import partial

import numpy as np

from grortir.main.model.core.abstract_stage import AbstractStage
from grortir.main.model.core.optimization_status import OptimizationStatus


def _obj_wrapper(func, args, kwargs, x):
    return func(x, *args, **kwargs)


def _is_feasible_wrapper(func, x):
    return np.all(func(x) >= 0)


def _cons_none_wrapper(x):
    return np.array([0])


def _cons_ieqcons_wrapper(ieqcons, args, kwargs, x):
    return np.array([y(x, *args, **kwargs) for y in ieqcons])


def _cons_f_ieqcons_wrapper(f_ieqcons, args, kwargs, x):
    return np.array(f_ieqcons(x, *args, **kwargs))


def _results(stage, *args):
    stage.control_params = args[0].tolist()
    if stage.is_enough_quality(args[1]):
        stage.optimization_status = OptimizationStatus.success
    else:
        stage.optimization_status = OptimizationStatus.failed
    return args


def pso(stage=AbstractStage(), ieqcons=[], f_ieqcons=None, args=(), kwargs={},
        swarmsize=100, omega=0.5, phip=0.5, phig=0.5, maxiter=100, minstep=1e-8,
        minfunc=1e-8, debug=False, processes=1, particle_output=False):
    """
    Perform a particle swarm optimization (PSO)

    Parameters
    ==========
    Optional
    ========
    stage : AbstractStage
        Stage which should be optimized.
    ieqcons : list
        A list of functions of length n such that ieqcons[j](x_positions_of_swarm,*args) >= 0.0 in
        a successfully optimized problem (Default: [])
    f_ieqcons : function
        Returns a 1-D array in which each element must be greater or equal
        to 0.0 in a successfully optimized problem. If f_ieqcons is specified,
        ieqcons is ignored (Default: None)
    args : tuple
        Additional arguments passed to objective and constraint functions
        (Default: empty tuple)
    kwargs : dict
        Additional keyword arguments passed to objective and constraint
        functions (Default: empty dict)
    swarmsize : int
        The number of particles in the swarm (Default: 100)
    omega : scalar
        Particle velocity scaling factor (Default: 0.5)
    phip : scalar
        Scaling factor to search away from the particle's best known position
        (Default: 0.5)
    phig : scalar
        Scaling factor to search away from the swarm's best known position
        (Default: 0.5)
    maxiter : int
        The maximum number of iterations for the swarm to search (Default: 100)
    minstep : scalar
        The minimum stepsize of swarm's best position before the search
        terminates (Default: 1e-8)
    minfunc : scalar
        The minimum change of swarm's best objective value before the search
        terminates (Default: 1e-8)
    debug : boolean
        If True, progress statements will be displayed every iteration
        (Default: False)
    processes : int
        The number of processes to use to evaluate objective function and
        constraints (default: 1)
    particle_output : boolean
        Whether to include the best per-particle position and the objective
        values at those.
    Returns
    =======
    g : array
        The swarm's best known position (optimal design)
    f : scalar
        The objective value at ``g``
    it : int
        Number of iterations
    p : array
        The best known position per particle
    pf: arrray
        The objective values at each position in p

    """
    func_to_opt = stage.get_quality
    stage.optimization_status = OptimizationStatus.in_progress
    lb = stage.lower_bounds
    ub = stage.upper_bounds
    assert len(lb) == len(ub), 'Lower- and upper-bounds must be the same length'
    assert hasattr(func_to_opt, '__call__'), 'Invalid function handle'
    lb = np.array(lb)
    ub = np.array(ub)
    assert np.all(
        ub > lb), 'All upper-bound values must be greater than lower-bound values'

    vhigh = np.abs(ub - lb)
    vlow = -vhigh

    # Initialize objective function
    obj = partial(_obj_wrapper, func_to_opt, args, kwargs)

    # Check for constraint function(s) #########################################
    if f_ieqcons is None:
        if not len(ieqcons):
            if debug:
                print('No constraints given.')
            cons = _cons_none_wrapper
        else:
            if debug:
                print('Converting ieqcons to a single constraint function')
            cons = partial(_cons_ieqcons_wrapper, ieqcons, args, kwargs)
    else:
        if debug:
            print('Single constraint function given in f_ieqcons')
        cons = partial(_cons_f_ieqcons_wrapper, f_ieqcons, args, kwargs)
    is_feasible = partial(_is_feasible_wrapper, cons)

    # Initialize the multiprocessing module if necessary
    if processes > 1:
        import multiprocessing
        mp_pool = multiprocessing.Pool(processes)

    # Initialize the particle swarm ############################################
    Swarm_size = swarmsize
    Dimensions = len(lb)  # the number of dimensions each particle has
    x_positions_of_swarm = np.random.rand(Swarm_size, Dimensions)  # particle positions
    velocity_of_particles = np.zeros_like(x_positions_of_swarm)  # particle velocities
    best_positions = np.zeros_like(x_positions_of_swarm)  # best particle positions
    current_func_values = np.zeros(Swarm_size)  # current particle function values
    feasibility = np.zeros(Swarm_size, dtype=bool)  # feasibility of each particle
    best_values = np.ones(Swarm_size) * np.inf  # best particle function values
    initial_best_position = np.inf  # best swarm position starting value

    # Initialize the particle's position
    x_positions_of_swarm = lb + x_positions_of_swarm * (ub - lb)

    # Calculate objective and constraints for each particle
    if processes > 1:
        current_func_values = np.array(mp_pool.map(obj, x_positions_of_swarm))
        feasibility = np.array(mp_pool.map(is_feasible, x_positions_of_swarm))
    else:
        for i in range(Swarm_size):
            current_func_values[i] = obj(x_positions_of_swarm[i, :])
            feasibility[i] = is_feasible(x_positions_of_swarm[i, :])

    # Store particle's best position (if constraints are satisfied)
    i_update = np.logical_and((current_func_values < best_values), feasibility)
    best_positions[i_update, :] = x_positions_of_swarm[i_update, :].copy()
    best_values[i_update] = current_func_values[i_update]

    # Update swarm's best position
    i_min = np.argmin(best_values)
    if best_values[i_min] < initial_best_position:
        initial_best_position = best_values[i_min]
        best_position = best_positions[i_min, :].copy()
    else:
        # At the start, there may not be any feasible starting point, so just
        # give it a temporary "best" point since it's likely to change
        best_position = x_positions_of_swarm[0, :].copy()

    # Initialize the particle's velocity
    velocity_of_particles = vlow + np.random.rand(Swarm_size, Dimensions) * (vhigh - vlow)

    # Iterate until termination criterion met ##################################
    it = 1
    while it <= maxiter and stage.could_be_optimized():
        rp = np.random.uniform(size=(Swarm_size, Dimensions))
        rg = np.random.uniform(size=(Swarm_size, Dimensions))

        # Update the particles velocities
        velocity_of_particles = omega * velocity_of_particles + phip * rp * (best_positions - x_positions_of_swarm) + phig * rg * (
            best_position - x_positions_of_swarm)
        # Update the particles' positions
        x_positions_of_swarm = x_positions_of_swarm + velocity_of_particles
        # Correct for bound violations
        maskl = x_positions_of_swarm < lb
        masku = x_positions_of_swarm > ub
        x_positions_of_swarm = x_positions_of_swarm * (~np.logical_or(maskl, masku)) + lb * maskl + ub * masku

        # Update objectives and constraints
        if processes > 1:
            current_func_values = np.array(mp_pool.map(obj, x_positions_of_swarm))
            feasibility = np.array(mp_pool.map(is_feasible, x_positions_of_swarm))
        else:
            for i in range(Swarm_size):
                current_func_values[i] = obj(x_positions_of_swarm[i, :])
                feasibility[i] = is_feasible(x_positions_of_swarm[i, :])

        # Store particle's best position (if constraints are satisfied)
        i_update = np.logical_and((current_func_values < best_values),
                                  feasibility)
        best_positions[i_update, :] = x_positions_of_swarm[i_update, :].copy()
        best_values[i_update] = current_func_values[i_update]

        # Compare swarm's best position with global best position
        i_min = np.argmin(best_values)
        if best_values[i_min] < initial_best_position:
            if debug:
                print('New best for swarm at iteration {:}: {:} {:}' \
                      .format(it, best_positions[i_min, :], best_values[i_min]))

            p_min = best_positions[i_min, :].copy()
            stepsize = np.sqrt(np.sum((best_position - p_min) ** 2))

            if np.abs(initial_best_position - best_values[i_min]) <= minfunc:
                print(
                    'Stopping search: Swarm best objective change less than {:}' \
                        .format(minfunc))
                return formated_values(particle_output, stage, best_position,
                                       initial_best_position, it,
                                       best_positions, best_values)
            elif stepsize <= minstep:
                print(
                    'Stopping search: Swarm best position change less than {:}' \
                        .format(minstep))
                return formated_values(particle_output, stage, best_position,
                                       initial_best_position, it,
                                       best_positions,
                                       best_values)
            else:
                best_position = p_min.copy()
                initial_best_position = best_values[i_min]

            if stage.is_enough_quality(best_values[i_min]):
                print('Stopping search: Enough quality reached.')
                return formated_values(particle_output, stage, best_position,
                                       initial_best_position, it,
                                       best_positions, best_values)

        if debug:
            print('Best after iteration {:}: {:} {:}'.format(it, best_position,
                                                             initial_best_position))
        it += 1
    if maxiter == it:
        print('Stopping search: maximum iterations reached --> {:}'.format(
            maxiter))
    else:
        print("Stopping search: stage couldn't be optimized")

    if not is_feasible(best_position):
        print(
            "However, the optimization couldn't find a feasible design. Sorry")
    return formated_values(particle_output, stage, best_position,
                           initial_best_position, it, best_positions,
                           best_values)


def formated_values(particle_output, stage, best_position,
                    initial_best_position,
                    it, best_positions, best_values):
    if particle_output:
        return _results(stage, best_position, initial_best_position, it,
                        best_positions, best_values)
    else:
        return _results(stage, best_position, initial_best_position, it)



"""Contains optimization controller."""
from grortir.main.model.core.optimization_status import OptimizationStatus


class OptimizationController(object):
    """Controller for optimization process."""

    @staticmethod
    def should_continue(stages):
        """Return true only if optimization should be continued."""
        for stage in stages:
            if OptimizationStatus.failed == stage.optimization_status:
                return False
        for stage in stages:
            if OptimizationStatus.success != stage.optimization_status:
                return True
        return False



"""Represents optimization strategy for PSO."""


class OptimizationStrategy:
    """Represents optimization strategy for PSO."""

    def get_group_optimization_strategy(self, stages_in_group):
        """
            Return optimization strategy for group.
        Args:
            stages_in_group (list): stages  in group

        Raises:
            NotImplementedError: when method not implemented.
        """
        raise NotImplementedError



"""Describe status of optimization."""
from enum import Enum


class OptimizationStatus(Enum):
    """Enum for describing status of optimization.
    not_started - nothing was optimized yet
    in_progress - optimization is in progress
    success - optimization successfully completed
    failed - optimization couldn't be done
    """
    not_started = "Not started"
    success = "Success"
    in_progress = "In progress"
    failed = "Failed"

    def __eq__(self, other):
        if other is None:
            return False
        return self.value == other.value



"""Module for testing optimizer."""

from unittest import TestCase
from unittest.mock import patch, Mock, sentinel

from grortir.main.model.core.abstract_process import AbstractProcess
from grortir.main.model.core.abstract_stage import AbstractStage
from grortir.main.model.core.optimization_status import OptimizationStatus
from grortir.main.optimizers.optimizer import Optimizer


class TestOptimizer(TestCase):
    """Class for testing Optimizer."""

    def setUp(self):
        """Set up environment."""
        self.some_process = AbstractProcess()
        self.first_stage = AbstractStage()
        self.second_stage = AbstractStage()
        self.some_process.add_path([self.first_stage, self.second_stage])

    def test___init__(self):
        """Testing creating object."""
        optimizer = Optimizer(self.some_process)
        self.assertIsNotNone(optimizer)
        self.assertEqual(optimizer.process, self.some_process)

    def test_set_order(self):
        """Test changing order of optimizing."""
        optimizer = Optimizer(self.some_process)
        optimizer.set_custom_optimizing_order(
            [self.second_stage, self.first_stage])
        self.assertEqual(optimizer.ordered_stages[0], self.second_stage)

    def test_set_order_fail(self):
        """Test changing order of optimizing with incorrect list."""
        optimizer = Optimizer(self.some_process)
        with self.assertRaises(ValueError):
            optimizer.set_custom_optimizing_order(
                [self.second_stage, self.first_stage, self.first_stage])

    @patch('grortir.main.optimizers.optimizer.pso')
    def test_optimize_process(self, pso_mock):
        """Test optimization of initial stage in process."""
        optimizer = Optimizer(self.some_process)
        result = optimizer.optimize_process()
        pso_mock.assert_called_once_with(self.first_stage, swarmsize=40)
        self.assertFalse(result)

    @patch('grortir.main.optimizers.optimizer.pso')
    def test_optimize_whole_process(self, pso_mock):
        """Test whole optimization."""
        optimizer = Optimizer(self.some_process)
        self.first_stage.optimization_status = OptimizationStatus.success
        self.first_stage.get_output_of_stage = Mock(return_value=sentinel)
        self.second_stage.optimization_status = OptimizationStatus.success
        result = optimizer.optimize_process()
        pso_mock.assert_called_with(self.second_stage, swarmsize=40)
        self.assertTrue(result)

    @patch('grortir.main.optimizers.optimizer.pso')
    def test_optimize_incorrect_process(self, pso_mock):
        """Test incorrect process structure."""
        optimizer = Optimizer(self.some_process)
        third_stage = AbstractStage()
        self.some_process.add_edge(third_stage, self.second_stage)
        self.first_stage.optimization_status = OptimizationStatus.success
        self.first_stage.get_output_of_stage = Mock(return_value=sentinel)
        pso_mock.assert_not_called()
        with self.assertRaises(AttributeError):
            optimizer.optimize_process()



from unittest import TestCase

from grortir.main.model.core.optimization_status import OptimizationStatus


class TestOptimizationStatus(TestCase):
    def test_equals_positive(self):
        self.assertTrue(OptimizationStatus.failed.__eq__(OptimizationStatus.failed))

    def test_equals_negative(self):
        self.assertFalse(
            OptimizationStatus.failed.__eq__(None))



"""A sample module."""


class Class(object):
    """A sample class."""


def function(value):
    """A function with branches to demonstrate branch coverage reporting."""
    if value is True:
        return 'True'
    elif value is False:
        return 'False'
    else:
        return 'None'


def function_with_network_stuff():
    """A sample function that might take a long time to test."""
    return True


def function_with_disk_stuff():
    """A sample function that might take a long time to test."""
    return False



"""Module contains class CallsStage."""
from grortir.main.model.core.abstract_stage import AbstractStage


# pylint: disable=arguments-differ,unused-argument


class CallsStage(AbstractStage):
    """Implementation of basic stage.

    Cost is calculated by number of calls of cost function.
    Attributes:
        cost (float): Actual cost of stage.
        max_calls (int): Maximum possible calls of cost quality function
        name (str): Name of stage
        maximum_acceptable_quality (float): max expected quality
    """

    def __init__(self, name, max_calls, input_vector=(),
                 maximum_acceptable_quality=0.01):
        """Constructor."""
        super().__init__(input_vector)
        self.max_calls = max_calls
        self.name = name
        self.control_params = [0] * len(self.input_vector)
        self.maximum_acceptable_quality = maximum_acceptable_quality
        self.cost = 0

    def get_quality(self, input_vector=None, control_params=None):
        """
        Return quality of actual output.

        Returns:
            float: quality

        """
        if control_params is None:
            control_params = self.control_params[:]
        self.cost += 1
        return self.calculate_quality(input_vector, control_params)

    @staticmethod
    def calculate_quality(input_vector, control_params):
        """
        Function for calculating quality.

        Returns:
            float: quality

        Raises:
            AssertionError: If length of `control_params`
                is not equal length of `input_vector`
        """
        assert len(control_params) == len(input_vector)
        quality = 0
        for i in enumerate(control_params):
            quality += (control_params[i[0]] - input_vector[
                i[0]]) ** 2
        return quality

    def get_cost(self):
        """
        Return actual cost of stage.

        Returns:
            float: cost
        """
        return self.cost

    def could_be_optimized(self):
        """Return answer if it is still possible to optimize that stage."""
        return self.get_cost() < self.get_maximal_acceptable_cost()

    def is_enough_quality(self, value):
        """Return True if value is proper quality."""
        return value <= self.maximum_acceptable_quality

    def get_output_of_stage(self, input_vector, control_params):
        """Return output of stage."""
        return control_params

    def get_maximal_acceptable_cost(self):
        """Return maximum acceptable cost."""
        return self.max_calls

    def __str__(self):
        return self.name



# pylint: disable=too-many-arguments


class CallRecord:
    def __init__(self, entity_name, method_name, input_parms_list, output,
                 count):
        self.entity_name = entity_name
        self.count = count
        self.output = output
        self.input_parms_list = input_parms_list
        self.method_name = method_name



"""Sample unit test module."""

import unittest

from grortir import sample


class TestGrortir(unittest.TestCase):

    """Sample unit test class."""

    def test_dependency_import(self):
        """Sample test method for dependencies."""
        try:
            import testpackage  # pylint: disable=unused-variable
            assert True
        except ImportError:
            self.fail("dependency not installed")

    def test_branch_coverage(self):
        """Sample test method for branch coverage."""
        self.assertEqual(sample.function(True), 'True')
        self.assertEqual(sample.function(False), 'False')
        self.assertEqual(sample.function(None), 'None')



"""Contain PsoAlgorithm."""
from grortir.main.model.core.optimization_status import OptimizationStatus
from grortir.main.pso.process_validator import ProcessValidator
from grortir.main.pso.whole_group_pso import WholeGroupPso


class PsoAlgorithm:
    """Optimize process with different strategies."""

    def __init__(self, process, grouping_strategy, optimization_startegy,
                 number_of_particle=40):
        self.process = process
        self.grouping_strategy = grouping_strategy
        self.optimization_strategy = optimization_startegy
        self.process_validator = ProcessValidator()
        self.whole_group_pso = WholeGroupPso(self.process, number_of_particle)

    def run(self):
        """Run algorithm."""
        self.process_validator.validate(self.process)
        self.process.optimizationStatus = OptimizationStatus.in_progress
        number_of_groups = self.grouping_strategy.get_actual_numbers_of_groups()
        for current_group_number in range(number_of_groups):
            current_stages = self.grouping_strategy.get_items_from_group(
                current_group_number)
            group_optimization_strategy = self.optimization_strategy. \
                get_group_optimization_strategy(current_stages)
            self.whole_group_pso.optimize(current_stages,
                                          group_optimization_strategy)



# pylint: skip-file
from grortir.main.model.processes.factories.calls_process_factory import \
    CallsProcessFactory
from grortir.main.optimizers.optimizer import Optimizer

five_stage_process = CallsProcessFactory("linear", 5, 1000,
                                         (0, 0, 0)).construct_process()
optimizer = Optimizer(five_stage_process)
optimizer.optimize_process()
print("end!")



"""Module for testing."""

from unittest import TestCase

from grortir.main.model.core.abstract_process import AbstractProcess


class TestAbstractProcess(TestCase):
    """Class for testing."""
    def test___init__(self):
        """Method which check constructor."""
        tested_object = AbstractProcess()
        self.assertIsNotNone(tested_object)



#!/usr/bin/env python

"""Setup script for Grortir."""

import setuptools

from grortir import __project__, __version__

try:
    README = open("README.rst").read()
    CHANGES = open("CHANGES.rst").read()
except IOError:
    DESCRIPTION = "<placeholder>"
else:
    DESCRIPTION = README + '\n' + CHANGES

setuptools.setup(
    name=__project__,
    version=__version__,

    description="""Application for supporting
                 optimization of production process.""",
    url='https://github.com/qbahn/grortir',
    author='Wojtek Pietrucha',
    author_email='wojtekpietrucha@gmail.com',

    packages=setuptools.find_packages(),

    entry_points={'console_scripts': []},

    long_description=(DESCRIPTION),
    license='MIT',
    classifiers=[
        'Development Status :: 1 - Planning',
        'Natural Language :: English',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
    ],

    install_requires=open("requirements.txt").readlines(),
)



from unittest import TestCase

from grortir.main.model.core.abstract_process import AbstractProcess
from grortir.main.model.core.abstract_stage import AbstractStage
from grortir.main.optimizers.base_optimizer import BaseOptimizer


class TestBaseOptimizer(TestCase):
    """Class for testing Optimizer."""

    def setUp(self):
        """Set up environment."""
        self.some_process = AbstractProcess()
        self.first_stage = AbstractStage()
        self.second_stage = AbstractStage()
        self.some_process.add_path([self.first_stage, self.second_stage])

    def test___init__(self):
        """Testing creating object."""
        optimizer = BaseOptimizer(self.some_process)
        self.assertIsNotNone(optimizer)
        self.assertEqual(optimizer.process, self.some_process)

    def test_set_order(self):
        """Test changing order of optimizing."""
        optimizer = BaseOptimizer(self.some_process)
        optimizer.set_custom_optimizing_order(
            [self.second_stage, self.first_stage])
        self.assertEqual(optimizer.ordered_stages[0], self.second_stage)

    def test_set_order_fail(self):
        """Test changing order of optimizing with incorrect list."""
        optimizer = BaseOptimizer(self.some_process)
        with self.assertRaises(ValueError):
            optimizer.set_custom_optimizing_order(
                [self.second_stage, self.first_stage, self.first_stage])

    def test_optimize_process(self):
        optimizer = BaseOptimizer(self.some_process)
        with self.assertRaises(NotImplementedError):
            optimizer.optimize_process()



"""Package for Grortir."""

import sys

__project__ = 'Grortir'
__version__ = '0.1.8'

VERSION = "{0} v{1}".format(__project__, __version__)

PYTHON_VERSION = 3, 3

if sys.version_info < PYTHON_VERSION:  # pragma: no cover (manual test)
    sys.exit("Python {0}.{1}+ is required.".format(*PYTHON_VERSION))



"""This test is checking that algorithm with grouping is working correctly."""
# pylint: disable=too-many-instance-attributes,too-many-arguments
# pylint: disable=signature-differs,arguments-differ

import unittest.mock as mock
from unittest import TestCase

from grortir.main.model.core.abstract_process import AbstractProcess
from grortir.main.optimizers.grouped_optimizer import GroupedOptimizer
from grortir.main.optimizers.grouping_strategy import GroupingStrategy
from grortir.main.model.core.abstract_stage import AbstractStage
from grortir.main.pso.calls_optimization_strategy import \
    CallsOptimizationStrategy
from grortir.main.pso.pso_algorithm import PsoAlgorithm
from tests.framework.calls.calls_registry import CallsRegistry
from tests.framework.calls.input_validator import InputValidator


class TestGroupedOptimizer(TestCase):
    """Class for testing Optimizer."""

    def setUp(self):
        """Set up environment."""
        self.some_process = AbstractProcess()
        self.first_stage = AbstractStage()
        self.second_stage = AbstractStage()
        self.third_stage_a = mock.Mock()
        self.third_stage_b = mock.Mock()
        self.some_process.add_path([self.first_stage, self.second_stage,
                                    self.third_stage_a])
        self.some_process.add_edge(self.second_stage, self.third_stage_b)

    def test___init__(self):
        """Testing creating object."""
        optimizer = GroupedOptimizer(self.some_process, mock.Mock(),
                                     mock.Mock())
        self.assertIsNotNone(optimizer)
        self.assertEqual(optimizer.process, self.some_process)

    def test_calling_functions(self):
        """Test correct order of calling function."""
        optimizer = GroupedOptimizer(TESTED_PROCESS, GROUPING_STRATEGY,
                                     PSO_ALGORITHM)
        optimizer.optimize_process()
        input_validator = InputValidator(CALLS_REGISTRY)
        is_valid = input_validator.validate_input(GROUPING_STRATEGY,
                                                  TESTED_PROCESS)
        self.assertTrue(is_valid)


CALLS_REGISTRY = CallsRegistry()


class DeterministicStage(AbstractStage):
    def __init__(self, name, max_get_output_of_stage_count=10,
                 max_is_enough_quality_count=10,
                 max_could_be_optimized_count=10, max_get_quality_count=10,
                 max_get_cost_count=10):
        super().__init__((0, 0, 0, 0))
        self.name = name
        self.get_cost_count = 0
        self.get_quality_count = 0
        self.could_be_optimized_count = 0
        self.is_enough_quality_count = 0
        self.get_output_of_stage_count = 0
        self.get_maximal_acceptable_cost_count = 0
        self.maximum_acceptable_quality = 1
        self.maximal_acceptable_cost = 100
        self.max_get_cost_count = max_get_cost_count
        self.max_get_quality_count = max_get_quality_count
        self.max_could_be_optimized_count = max_could_be_optimized_count
        self.max_is_enough_quality_count = max_is_enough_quality_count
        self.max_get_output_of_stage_count = max_get_output_of_stage_count
        CALLS_REGISTRY.add_call(self.name, '__init__',
                                [name, max_get_output_of_stage_count,
                                 max_is_enough_quality_count,
                                 max_could_be_optimized_count,
                                 max_get_quality_count,
                                 max_get_cost_count], None, 1)

    def get_cost(self):
        self.get_cost_count += 1
        CALLS_REGISTRY.add_call(self.name, "get_cost", [], self.get_cost_count,
                                self.get_cost_count)
        return self.get_cost_count

    def get_quality(self, input_vector, control_params=None):
        self.get_quality_count += 1
        result = 0.1 if self.get_quality_count > int(self.name) else 1000
        CALLS_REGISTRY.add_call(self.name, "get_quality",
                                [input_vector, control_params], result,
                                self.get_quality_count)
        return result

    def could_be_optimized(self):
        self.could_be_optimized_count += 1
        CALLS_REGISTRY.add_call(self.name, "could_be_optimized", [], True,
                                self.could_be_optimized_count)
        return True

    def is_enough_quality(self, value):
        self.is_enough_quality_count += 1
        result = value < 1
        CALLS_REGISTRY.add_call(self.name, "is_enough_quality", [value], result,
                                self.is_enough_quality_count)
        return result

    def get_output_of_stage(self, input_vector, control_params):
        self.get_output_of_stage_count += 1
        output = [self.get_output_of_stage_count] * len(input_vector)
        output[0] = int(self.name)
        CALLS_REGISTRY.add_call(self.name, "get_output_of_stage",
                                [input_vector, control_params], output,
                                self.get_output_of_stage_count)
        return output

    def get_maximal_acceptable_cost(self):
        self.get_maximal_acceptable_cost_count += 1
        CALLS_REGISTRY.add_call(self.name, "get_maximal_acceptable_cost",
                                [], self.maximal_acceptable_cost,
                                self.get_output_of_stage_count)
        return self.maximal_acceptable_cost


class ExampleProcess(AbstractProcess):
    pass


TESTED_PROCESS = ExampleProcess()
stages = {}
for i in range(7):
    stages[i] = DeterministicStage(str(i))

# Our graph:
#   0
#   |
#   1
#   |\
#   2 4
#   | |\
#   3 5 6
# All edges directed to down
# Order of nodes is the same as names
TESTED_PROCESS.add_edge(stages[0], stages[1])
TESTED_PROCESS.add_edge(stages[1], stages[2])
TESTED_PROCESS.add_edge(stages[2], stages[3])

TESTED_PROCESS.add_edge(stages[1], stages[4])
TESTED_PROCESS.add_edge(stages[4], stages[5])
TESTED_PROCESS.add_edge(stages[4], stages[6])

# Groups:
#   (0)0
#      |
#   (0)1
#      |\
#   (0)2 4(1)
#      | | \
#   (1)3 5(1)6(2)

GROUPING_STRATEGY = GroupingStrategy(list(stages.values()))
GROUPING_STRATEGY.define_group((stages[0], stages[1], stages[2]))
GROUPING_STRATEGY.define_group((stages[3], stages[4], stages[5]))
GROUPING_STRATEGY.define_group((stages[6],))

OPTIMIZATION_STARTEGY = CallsOptimizationStrategy()

PSO_ALGORITHM = PsoAlgorithm(TESTED_PROCESS, GROUPING_STRATEGY,
                             OPTIMIZATION_STARTEGY, 2)



"""Base optimizer."""
import networkx as nx


class BaseOptimizer(object):
    """Optimizer is object which optimize process."""

    def __init__(self, process):
        self.process = process
        self.ordered_stages = nx.topological_sort(self.process)
        self.swarm_size = 40

    def set_custom_optimizing_order(self, ordered_stages):
        """Set custom order.

        Raises:
            ValueError: When order doesn't contain all stages.
        """
        if set(self.ordered_stages) == set(ordered_stages) and len(
                self.ordered_stages) == len(ordered_stages):
            self.ordered_stages = ordered_stages
        else:
            raise ValueError("List of stages must contain all stages.")

    def optimize_process(self):
        """Optimize process.

        Raises:
            NotImplementedError: Abstract method.
        """
        raise NotImplementedError



"""Represents optimization strategy for group in PSO."""


class GroupOptimizationStrategy:
    """Represents optimization strategy for group in PSO."""

    @staticmethod
    def initialize():
        """
        Initialize strategy.

        Raises:
            NotImplementedError: if not  implemented.
        """
        raise NotImplementedError

    @staticmethod
    def should_continue(best_particle):
        """
        Strategic method.

        Args:
            best_particle (Particle): best particle in swarm

        Raises:
            NotImplementedError: if not implemented.
        '"""
        raise NotImplementedError
